#include "i0mmonplot.h"
#include <cumbiapool.h>
#include <cucontrolsfactorypool.h>
#include <quplotcurve.h>
#include <qutimescaledraw.h>
#include <qwt_symbol.h>
#include <qwt_date_scale_engine.h>

#include <cupluginloader.h>
#include <qumultireaderplugininterface.h>
#include <cumacros.h>

#include <QtDebug>

#ifdef QUMBIA_TANGO_CONTROLS_VERSION
#include <QRegularExpression>
#include <cutreader.h>
#endif

I0MMonPlot::I0MMonPlot(QWidget *parent, CumbiaPool *cup, const CuControlsFactoryPool &fpoo)
    : QuPlotBase(parent),
      m_symstyles (QList<QwtSymbol::Style> () <<  QwtSymbol::XCross << QwtSymbol::Cross
                                               << QwtSymbol::Rect << QwtSymbol::Triangle << QwtSymbol::Star1 << QwtSymbol::Diamond << QwtSymbol::Ellipse) {
    m_ctrl_fpool = fpoo;
    m_cu_poo = cup;
    m_time_mode = m_empty = true;
    m_numsamples = 1;
    m_ymin = m_ymax = 0;
    CuPluginLoader plo;
    QObject *plugin_qob;
    m_multiread_plu_i = plo.get<QuMultiReaderPluginInterface>("libcumbia-multiread-plugin.so", &plugin_qob);
    if(!m_multiread_plu_i)
        perr("I0MMonPlot: failed to load plugin \"libcumbia-multiread-plugin.so\"");
    else {
        int mode = 0;
#ifdef QUMBIA_TANGO_CONTROLS_VERSION
        mode = CuTReader::Manual;
#endif
        m_multiread_plu_i->init(cup, fpoo, mode); // 0 sync mode
        connect(m_multiread_plu_i->get_qobject(), SIGNAL(onSeqReadComplete(const QList<CuData> &)),
                this, SLOT(onDataReady(const QList<CuData> &)));
    }

    setXAxisAutoscaleEnabled(false);
    setYAxisAutoscaleEnabled(false);
}

bool I0MMonPlot::pluginLoaded() const {
    return m_multiread_plu_i != nullptr;
}

bool I0MMonPlot::yAutoScale() const {
    return m_ymax == m_ymin;
}
void I0MMonPlot::setSources(const QString &x, const QString &y, const QMap<int, double> lambdamap, int nsamples)
{
    qDebug() << __PRETTY_FUNCTION__ << x << y << lambdamap;
    QuPlotBase::clearPlot();
    m_multiread_plu_i->unsetSources();
    m_curvesymap.clear();
    m_empty = true;
    m_numsamples = nsamples;
    m_lambdas = lambdamap;
    m_time_mode = (x == "time");
    m_reset_scales();
    foreach(double l, lambdamap.values()) {
        QString cuna = QString("lambda %1").arg(l);
        QuPlotCurve *crv = curve(cuna);
        if(!crv) {
            crv = new QuPlotCurve(cuna);
            addCurve(cuna, crv);
        }
        m_curvesymap[cuna] = m_symstyles[m_curvesymap.size() % m_symstyles.size()];
    }
    if(m_time_mode) {
        if(m_multiread_plu_i)
            m_multiread_plu_i->insertSource(y, 0);
    }
    else {
        if(m_multiread_plu_i ) {
            QString xsrc;
            x.contains(QRegularExpression("\\(\\d,%\\d\\)")) ? xsrc = QString(x).arg(nsamples) : xsrc = x;
            m_multiread_plu_i->insertSource(xsrc, 0);
            m_multiread_plu_i->insertSource(y, 1);
        }
    }
}

void I0MMonPlot::setYBounds(double ymi, double yma) {
    if(ymi != yma && ymi < yma)  {
        m_ymin = ymi;
        m_ymax = yma;
        setYLowerBound(ymi);
        setYUpperBound(yma);
    }
}

void I0MMonPlot::setYBoundsAuto(bool a) {
    if(a) {
        m_ymin = m_ymax = 0;
        m_empty = true;
    }
}

void I0MMonPlot::m_set_bounds(const QVector<double> &xv, const QVector<double> &yv)
{
    if(m_ymin == m_ymax) { // auto scale Y
        double ylb, yub;
        const auto [ym, yM] = std::minmax_element(begin(yv), end(yv));
                ylb = m_empty ? (*ym) : yLowerBound();
                yub = m_empty ? (*yM) : yUpperBound();
                // y auto scale?
                if(*yM == *ym) {
            if(ylb >= *ym)
                ylb = (*ym - 0.5);
            if(yub <= *yM)
                yub = (*yM + 0.5);
        } else {
            if(ylb >= *ym)
                ylb = (*ym - 0.15 * (*yM - *ym));
            if(yub <= *yM)
                yub = (*yM + 0.15 * (*yM - *ym));
        }

        if(ylb != yLowerBound())
            setYLowerBound(ylb);
        if(yub != yUpperBound())
            setYUpperBound(yub);
    }

    const auto [xm, xM] = std::minmax_element(begin(xv), end(xv));
    double xlb, xub;
    xlb = m_empty ? (*xm) : xLowerBound();
    xub = m_empty ? (*xM) : xUpperBound();

    if(*xm == *xM) {
        if(xlb >= *xm)
            xlb = (*xm - 0.5);
        if(xub <= *xM)
            xub = (*xM + 0.5);
    } else {
        if(xlb >= *xm)
            xlb = (*xm - 0.15 * (*xM - *xm));
        if(xub <= *xM)
            xub = (*xM + 0.15 * (*xM - *xm));
    }

    if(xlb != xLowerBound())
        setXLowerBound(xlb);
    if(xub != xUpperBound())
        setXUpperBound(xub);
}


void I0MMonPlot::onDataReady(const QList<CuData> &dl) {
    qDebug() << __PRETTY_FUNCTION__ << "data szi" <<  dl.size();
    double x = 0.0;

    if(!m_time_mode && dl.size() == 2) { // correlate
        QVector<double> yv, xv;
        const CuData& xd = dl[0];
        const CuData& yd = dl[1];
        printf("\e[1;35mx %s , y %s\e[0m\n", vtoc2(xd, "value"), vtoc2(yd, "value"));
        if(xd.has("data_format_str", "vector")) {
            // average
            double sum = 0.0;
            const std::vector<double> &v = xd["value"].toDoubleVector();
            for(size_t i = 0; i < v.size(); i++)
                sum += v[i];
            x = sum / m_numsamples;
            printf("x (%f) averaged across %ld samples\n", x, v.size());
        }
        else
            x = xd["value"].toDouble();
        // y
        const std::vector<double> &yt = yd["value"].toDoubleVector();
        foreach(int i, m_lambdas.keys()) {
            if(i < static_cast<int>(yt.size()) ) {
                appendData(QString("lambda %1").arg(m_lambdas[i]), x, yt[i]);
                yv << yt[i];
                xv << x;
            }
        }
        m_set_bounds(xv, yv);
    }
    else if(m_time_mode) { // trend in time

        const CuData &da = dl[0];
        const std::vector<double> &ystdv = da["value"].toDoubleVector();
        QVector<double> xv {0.0}; // m_set_bounds needs a vector, say 0 and data size
#if QT_VERSION < QT_VERSION_CHECK(5, 12, 0)
        const QVector<double> yv = QVector<double>::fromStdVector(ystdv);
#else
        const QVector<double> yv(ystdv.begin(), ystdv.end());
#endif
        foreach(int idx, m_lambdas.keys()) {
            const QString curveName = QString("lambda %1").arg(m_lambdas[idx]);
            const QuPlotCurve *c = curve(curveName);
            xv << c->data()->size();
            if(idx < static_cast<int>(yv.size()) && curve(curveName) != nullptr ) {
                m_set_bounds(xv, yv);
                appendData(curveName, xv[1], yv[idx]);
            }
        }
    }
    else {
        printf("\e[1;31mI0MMonPlot::onDataReady: no time mode: data size %d\e[0m\n", dl.size());
        for(int i = 0; i < dl.size(); i++)
            printf("\e[1;31mI0MMonPlot::onDataReady: %s\e[0m\n", datos(dl[i]));
    }
    m_curves_style();
    m_empty = false;
}

void I0MMonPlot::m_reset_scales() {
    setXLowerBound(-0.5);
    setXUpperBound(0.5);
    if(m_ymin == m_ymax) {
        setYLowerBound(-0.5);
        setYUpperBound(0.5);
    }
}

void I0MMonPlot::m_curves_style() {
    foreach(QString cn, m_curvesymap.keys()) {
        QuPlotCurve *c = curve(cn);
        if(c && c->style() != QwtPlotCurve::Dots) {
            c->setStyle(QwtPlotCurve::Dots);
            //            QPen p = c->pen();
            //            p.setWidthF(2.0f);
            //            c->setPen(p);
        }
#if CUMBIA_QTCONTROLS_VERSION > 0x010202
        c->setMaxDataSizeForSymbol(0); // disable first points symbol
#endif
        if(!c->symbol() || c->symbol()->style() != m_curvesymap[cn])
            c->setSymbol(new QwtSymbol(m_curvesymap[cn], c->brush(), c->pen(), QSize(6, 6)));
    }
}

